import functools
import hashlib
import json
import os
import shutil
import socket
import tempfile
import zipfile
from ckan import model
from ckan.lib import mailer
from ckan.plugins import toolkit, PluginImplementations
from datetime import datetime
from splitgill.indexing.utils import get_elasticsearch_client
from splitgill.search import create_version_query
from elasticsearch_dsl import Search
from glob import iglob
from jinja2 import Template
from traceback import format_exception_only

from .dwc import dwc_writer
from .jsonl import jsonl_writer
from .sv import sv_writer
from .xlsx import xlsx_writer
from .utils import calculate_field_counts
from .transform import Transform
from .. import common
from ..datastore_utils import trim_index_name, prefix_resource
from ...interfaces import IVersionedDatastoreDownloads
from ...model.downloads import (
    DatastoreDownload,
    state_complete,
    state_failed,
    state_processing,
    state_zipping,
)

format_registry = {
    'csv': functools.partial(sv_writer, dialect='excel', extension='csv'),
    'tsv': functools.partial(sv_writer, dialect='excel-tab', extension='tsv'),
    'jsonl': jsonl_writer,
    'dwc': dwc_writer,
    'xlsx': xlsx_writer,
}

default_body = '''
Hello,
The link to the resource data you requested on {{ site_url }} is available at {{ download_url }}.

Best Wishes,
The Download Bot
'''.strip()

default_html_body = '''
<html lang="en">
<body>
<p>Hello,</p>
<p>The link to the resource data you requested on <a href="{{ site_url }}">{{ site_url }}</a> is
available at <a href="{{ download_url }}">here</a>.</p>
<br />
<p>Best Wishes,</p>
<p>The Download Bot</p>
</body>
</html>
'''.strip()


def queue_download(
    email_address,
    download_id,
    query_hash,
    query,
    query_version,
    search,
    resource_ids_and_versions,
    separate_files,
    file_format,
    format_args,
    ignore_empty_fields,
    transform,
):
    """
    Queues a job which when run will download the data for the resource.

    :return: the queued job
    """
    request = DownloadRequest(
        email_address,
        download_id,
        query_hash,
        query,
        query_version,
        search,
        resource_ids_and_versions,
        separate_files,
        file_format,
        format_args,
        ignore_empty_fields,
        transform,
    )
    # pass a timeout of 1 hour (3600 seconds)
    return toolkit.enqueue_job(
        download,
        args=[request],
        queue='download',
        title=str(request),
        rq_kwargs={'timeout': 3600},
    )


class DownloadRequest(object):
    """
    Class representing a request to download data.

    We use a class like this for two reasons, firstly to avoid having a long list of
    arguments passed through to queued functions, and secondly because rq by default
    logs the arguments sent to a function and if the records argument is a large list of
    dicts this becomes insane.
    """

    def __init__(
        self,
        email_address,
        download_id,
        query_hash,
        query,
        query_version,
        search,
        resource_ids_and_versions,
        separate_files,
        file_format,
        format_args,
        ignore_empty_fields,
        transform,
    ):
        self.email_address = email_address
        self.download_id = download_id
        self.query_hash = query_hash
        self.query = query
        self.query_version = query_version
        self.search = search
        self.resource_ids_and_versions = resource_ids_and_versions
        self.separate_files = separate_files
        self.file_format = file_format
        self.format_args = format_args
        self.ignore_empty_fields = ignore_empty_fields
        self.transform = transform

    @property
    def resource_ids(self):
        return sorted(self.resource_ids_and_versions.keys())

    def update_download(self, **kwargs):
        """
        Update the DatastoreDownload object with the given database id with the given
        update dict. The update dict will be passed directly to SQLAlchemy.

        :param kwargs: a dict of fields to update with the corresponding values
        """
        download_entry = model.Session.query(DatastoreDownload).get(self.download_id)
        for field, value in kwargs.items():
            setattr(download_entry, field, value)
        download_entry.save()

    def generate_download_hash(self):
        to_hash = [
            self.query_hash,
            self.query_version,
            sorted(self.resource_ids_and_versions.items()),
            self.separate_files,
            self.file_format,
            self.format_args,
            self.ignore_empty_fields,
            self.transform,
        ]
        download_hash = hashlib.sha1('|'.join(map(str, to_hash)).encode('utf-8'))
        return download_hash.hexdigest()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'Download data, id: {self.download_id}'


def download(request):
    """
    Downloads the data found by the query in the request and writes it to a zip.

    :param request: the DownloadRequest object, see that for parameter details
    """
    download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
    # make sure this download dir exists
    if not os.path.exists(download_dir):
        os.mkdir(download_dir)
    download_hash = request.generate_download_hash()
    existing_file = next(
        iglob(os.path.join(download_dir, f'*_{download_hash}.zip')), None
    )
    if existing_file is not None:
        previous_download_id = os.path.split(existing_file)[1].split('_')[0]
        existing_download = model.Session.query(DatastoreDownload).get(
            previous_download_id
        )
        if existing_download is not None:
            request.update_download(
                state='complete',
                total=existing_download.total,
                resource_totals=existing_download.resource_totals,
            )
            zip_name = f'{existing_download.id}_{download_hash}.zip'
            send_email(request, zip_name)
            return

    zip_name = f'{request.download_id}_{download_hash}.zip'
    zip_path = os.path.join(download_dir, zip_name)

    start = datetime.now()
    request.update_download(state=state_processing)
    target_dir = tempfile.mkdtemp()

    try:
        es_client = get_elasticsearch_client(
            common.CONFIG,
            sniff_on_start=True,
            sniffer_timeout=60,
            sniff_on_connection_fail=True,
            sniff_timeout=10,
            http_compress=False,
            timeout=30,
        )

        for plugin in PluginImplementations(IVersionedDatastoreDownloads):
            request = plugin.download_before_write(request)

        # this manifest will be written out as JSON and put in the download zip
        manifest = {
            'download_id': request.download_id,
            'resources': {},
            'separate_files': request.separate_files,
            'file_format': request.file_format,
            'format_args': request.format_args,
            'ignore_empty_fields': request.ignore_empty_fields,
            'transform': request.transform,
        }
        # calculate, per resource, the number of values for each field present in the search
        field_counts = calculate_field_counts(request, es_client)
        # choose the writer function based on the requested file format
        writer_function = format_registry[request.file_format]
        # we shouldn't auth anything
        context = {'ignore_auth': True}
        # keep track of the resource record counts
        resource_counts = {}

        with writer_function(request, target_dir, field_counts) as writer:
            # handle each resource individually. We could search across all resources at the same
            # but we don't need to seeing as we're not doing sorting here. By handling each index
            # one at a time we should be less of an es server burden and also it means we can use
            # the scroll api - if we were to search all resources requested at once we'd probably
            # have to use the multisearch endpoint on es and that doesn't support the scroll api
            for resource_id, version in request.resource_ids_and_versions.items():
                search = (
                    Search.from_dict(request.search)
                    .index(prefix_resource(resource_id))
                    .using(es_client)
                    .filter(create_version_query(version))
                )

                total_records = 0
                for hit in search.scan():
                    data = hit.data.to_dict()
                    data = Transform.transform_data(data, request.transform)
                    resource_id = trim_index_name(hit.meta.index)
                    # call the write function returned by our format specific writer context manager
                    writer(hit, data, resource_id)
                    total_records += 1

                # if no records were written, move on and don't populate the manifest
                if total_records == 0:
                    continue

                # retrieve information about the resource and package and add it to the manifest
                resource_dict = toolkit.get_action('resource_show')(
                    context, {'id': resource_id}
                )
                package_id = resource_dict['package_id']
                package_dict = toolkit.get_action('package_show')(
                    context, {'id': package_id}
                )
                manifest['resources'][resource_id] = {
                    'name': resource_dict['name'],
                    'package_id': package_id,
                    'package_title': package_dict['title'],
                    'total_records': total_records,
                    'field_counts': field_counts[resource_id],
                    'version': version,
                }
                resource_counts[resource_id] = total_records

        overall_total = sum(resource_counts.values())
        request.update_download(
            state=state_zipping, total=overall_total, resource_totals=resource_counts
        )

        # create a list of files that should be added to the zip, this should include the manifest
        files_to_zip = ['manifest.json'] + os.listdir(target_dir)

        # add the final data to the manifest
        manifest['total_records'] = overall_total
        manifest['files'] = files_to_zip
        end = datetime.now()
        manifest['start'] = start.isoformat()
        manifest['end'] = end.isoformat()
        manifest['duration_in_seconds'] = (end - start).total_seconds()

        # write the manifest out
        with open(os.path.join(target_dir, 'manifest.json'), 'w', encoding='utf8') as f:
            json.dump(manifest, f, sort_keys=True, indent=2, ensure_ascii=False)

        # zip up the files into the downloads directory
        with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED, True) as z:
            for filename in files_to_zip:
                z.write(os.path.join(target_dir, filename), arcname=filename)

        # drop the user an email to let them know their download is ready
        try:
            send_email(request, zip_name)
        except (mailer.MailerException, socket.error):
            raise

        request.update_download(state=state_complete)
    except Exception as error:
        error_message = str(format_exception_only(type(error), error)[-1].strip())
        request.update_download(state=state_failed, error=error_message)
        raise
    finally:
        # remove the temp dir we were using
        shutil.rmtree(target_dir)


def send_email(request, zip_name):
    """
    Sends an email to the email address in the passed request informing them that a
    download has completed and providing them with a link to go get it from.

    :param request: the DownloadRequest object
    :param zip_name: the name of the zip file that has been created
    """
    # get the templates as strings
    templates = (default_body, default_html_body)
    for plugin in PluginImplementations(IVersionedDatastoreDownloads):
        templates = plugin.download_modify_email_templates(*templates)

    site_url = toolkit.config.get('ckan.site_url')
    context = {
        'site_url': site_url,
        'download_url': f'{site_url}/downloads/{zip_name}',
    }

    # allow plugins to modify the context passed to the templates
    for plugin in PluginImplementations(IVersionedDatastoreDownloads):
        context = plugin.download_modify_email_template_context(request, context)

    # render both templates
    body, body_html = (Template(template).render(**context) for template in templates)

    # vend
    mailer.mail_recipient(
        recipient_email=request.email_address,
        recipient_name='Downloader',
        subject='Data download',
        body=body,
        body_html=body_html,
    )



import contextlib
from collections import defaultdict

from ckan.plugins import toolkit

from . import urls
from .archive import Archive
from .schema import Schema
from ..utils import filter_data_fields, get_fields


@contextlib.contextmanager
def dwc_writer(request, target_dir, field_counts):
    """
    Provides the functionality to write data to DarwinCore archives
    (https://dwc.tdwg.org).

    This function is a context manager and when used with `with` will yield a write function. This
    function takes 3 parameters - the elasticsearch hit object, the data dict from the hit and the
    resource id the hit came from.

    :param request: the download request object
    :param target_dir: the target directory to put the data files in
    :param field_counts: a dict of resource ids -> fields -> counts used to determine which fields
                         should be included in the csv file headers
    :return: yields a write function
    """

    # load the extensions, either from ckan.ini or from request args (request args take precedence).
    # Extensions are specified by name (as in urls.py).
    schema_args = {}
    # there can only be one core extension
    core_ext = request.format_args.get(
        'core_extension_name',
        toolkit.config.get('ckanext.versioned_datastore.dwc_core_extension_name'),
    )
    if core_ext is not None and core_ext.lower() in urls.core_extensions:
        core_ext_url = urls.core_extensions.get(core_ext.lower())
        schema_args['core_extension_url'] = core_ext_url

    # there can be multiple non-core extensions, separated by commas in ckan.ini or supplied as a
    # list in the request args
    ext_names = request.format_args.get(
        'extension_names',
        [
            e.strip().lower()
            for e in toolkit.config.get(
                'ckanext.versioned_datastore.dwc_extension_names', ''
            ).split(',')
        ],
    )
    # the fields used by the extension can also be overridden by request args
    config_extension_map = request.format_args.get('extension_map', {})
    ext_urls = []
    for e in ext_names:
        ext = urls.extensions.get(e)
        if not ext:
            continue
        fields = config_extension_map.get(e)
        if fields is not None and isinstance(fields, list):
            ext.fields = fields
        if ext.fields:  # no point in adding it if no fields are defined
            ext_urls.append(ext)

    if len(ext_urls) > 0:
        schema_args['extension_urls'] = ext_urls
    schema_controller = Schema.load(
        toolkit.config.get('ckanext.versioned_datastore.dwc_schema_cache'),
        **schema_args
    )

    if request.separate_files:
        # files will be opened lazily and stored in this dict using the resource ids as keys
        open_files = {}
    else:
        all_field_names = get_fields(field_counts, request.ignore_empty_fields)
        # each 'Archive' contains multiple open files
        open_file = Archive(schema_controller, request, target_dir).open(
            all_field_names
        )
        # ensure that any request to get the open file for a given resource returns this archive
        open_files = defaultdict(lambda: open_file)

    try:

        def write(hit, data, resource_id):
            if request.separate_files and resource_id not in open_files:
                resource_field_names = get_fields(
                    field_counts, request.ignore_empty_fields, [resource_id]
                )
                archive = Archive(
                    schema_controller, request, target_dir, resource_id
                ).open(resource_field_names)
                # lazily open the archive for this resource id
                open_files[resource_id] = archive
            else:
                archive = open_files[resource_id]

            if request.ignore_empty_fields:
                data = filter_data_fields(data, field_counts[resource_id])

            archive.initialise(data)

            # always add the resource ID
            data['datasetID'] = resource_id

            # and the record type
            data['basisOfRecord'] = archive.schema.row_type_name

            # write the data
            archive.write_record(data)

        yield write
    finally:
        # make sure we close all the files we opened
        for open_file in open_files.values():
            open_file.close()



from datetime import datetime

from ckan.plugins import toolkit
from splitgill.utils import to_timestamp

from .meta import help, schema
from ckantools.decorators import action
from ...lib import common
from ...lib.datastore_utils import prefix_resource
from ...lib.downloads.download import queue_download
from ...lib.query.schema import (
    get_latest_query_version,
    validate_query,
    translate_query,
    hash_query,
)
from ...lib.query.utils import get_available_datastore_resources
from ...model.downloads import DatastoreDownload


@action(schema.datastore_queue_download(), help.datastore_queue_download)
def datastore_queue_download(
    email_address,
    context,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
    separate_files=True,
    format='csv',
    format_args=None,
    ignore_empty_fields=True,
    transform=None,
    slug_or_doi=None,
):
    """
    Starts a download of the data found by the given query parameters. This download is
    created asynchronously using the rq background job queue and a link to the results
    is emailed to the given email address when complete.

    :param email_address: the email address to send the download link to
    :param context: the context dict from the action call
    :param query: the query dict. If None (default) then an empty query is used
    :param query_version: the version of the query schema the query is using. If None (default) then
                          the latest query schema version is used
    :param version: the version to search the data at. If None (default) the current time is used
    :param resource_ids: the list of resource to search. If None (default) then all the resources
                         the user has access to are queried. If a list of resources are passed then
                         any resources not accessible to the user will be removed before querying
    :param resource_ids_and_versions: a dict of resources and versions to search each of them at.
                                      This allows precise searching of each resource at a specific
                                      parameter. If None (default) then the resource_ids parameter
                                      is used together with the version parameter. If this parameter
                                      is provided though, it takes priority over the resource_ids
                                      and version parameters.
    :param separate_files: whether to split the results into a file per resource or just put all
                           results in one file. The default is True - split results into a file per
                           resource.
    :param format: the format to download the data in. The default is csv.
    :param format_args: additional arguments for specific formats, e.g. extension names for DwC-A.
    :param ignore_empty_fields: whether to ignore fields with no data in them in the result set
                                and not write them into the download file(s). Default: True.
    :param transform: data transformation configuration.
    :param slug_or_doi: use instead of query, query_version, resource_ids, and resource_ids_and_versions;
                        retrieves these parameters from a saved query via a slug or a doi
    :return: a dict containing info about the background job that is doing the downloading and the
             download id
    """

    if slug_or_doi:
        try:
            saved_query = toolkit.get_action('datastore_resolve_slug')(
                context, {'slug': slug_or_doi}
            )
            query = saved_query.get('query')
            query_version = saved_query.get('query_version')
            resource_ids = saved_query.get('resource_ids')
            resource_ids_and_versions = saved_query.get('resource_ids_and_versions')
        except toolkit.ValidationError:
            # if the slug doesn't resolve, continue as normal
            pass

    if resource_ids_and_versions is None:
        resource_ids_and_versions = {}
    else:
        # use the resource_ids_and_versions dict first over the resource_ids and version params
        resource_ids = list(resource_ids_and_versions.keys())

    # figure out which resources should be searched
    resource_ids = get_available_datastore_resources(context, resource_ids)
    if not resource_ids:
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    rounded_resource_ids_and_versions = {}
    # see if a version was provided, we'll use this is a resource id we're searching doesn't have a
    # directly assigned version (i.e. it was absent from the requested_resource_ids_and_versions
    # dict, or that parameter wasn't provided)
    if version is None:
        version = to_timestamp(datetime.now())
    for resource_id in resource_ids:
        # try to get the target version from the passed resource_ids_and_versions dict, but if
        # it's not in there, default to the version variable
        target_version = resource_ids_and_versions.get(resource_id, version)
        index = prefix_resource(resource_id)
        # round the version down to ensure we search the exact version requested
        rounded_version = common.SEARCH_HELPER.get_rounded_versions(
            [index], target_version
        )[index]
        if rounded_version is not None:
            # resource ids without a rounded version are skipped
            rounded_resource_ids_and_versions[resource_id] = rounded_version

    # setup the query
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()
    validate_query(query, query_version)
    search = translate_query(query, query_version)
    query_hash = hash_query(query, query_version)

    format_args = format_args or {}
    transform = transform or {}

    options = {
        'separate_files': separate_files,
        'format': format,
        'format_args': format_args,
        'ignore_empty_fields': ignore_empty_fields,
        'transform': transform,
    }
    download = DatastoreDownload(
        query_hash=query_hash,
        query=query,
        query_version=query_version,
        resource_ids_and_versions=rounded_resource_ids_and_versions,
        state='queued',
        options=options,
    )
    download.save()

    job = queue_download(
        email_address,
        download.id,
        query_hash,
        query,
        query_version,
        search.to_dict(),
        rounded_resource_ids_and_versions,
        separate_files,
        format,
        format_args,
        ignore_empty_fields,
        transform,
    )

    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
        'download_id': download.id,
    }



import json
import os
from collections import defaultdict

import contextlib

from .utils import filter_data_fields


@contextlib.contextmanager
def jsonl_writer(request, target_dir, field_counts):
    """
    Provides the functionality to write data to jsonl files (JSON lines:
    http://jsonlines.org/). Each line in the file is a JSON serialised representation of
    a record. This function handles opening and closing the necessary files to write
    data to either one file or a file per resource.

    This function is a context manager and when used with `with` will yield a write function. This
    function takes 3 parameters - the elasticsearch hit object, the data dict from the hit and the
    resource id the hit came from.

    :param request: the download request object
    :param target_dir: the target directory to put the data files in
    :param field_counts: not used, just here to match the writer interface
    :return: yields a write function
    """
    if request.separate_files:
        # files will be opened lazily and stored in this dict using the resource ids as keys
        open_files = {}
    else:
        # open the single file we're going to write to
        open_file = open(os.path.join(target_dir, 'data.jsonl'), 'w', encoding='utf-8')
        # ensure that any request to get the open file for a given resource always returns the one
        # file we've opened
        open_files = defaultdict(lambda: open_file)

    try:

        def write(hit, data, resource_id):
            if request.separate_files and resource_id not in open_files:
                # lazily open the file for this resource id
                resource_file_name = os.path.join(target_dir, f'{resource_id}.jsonl')
                open_files[resource_id] = open(
                    resource_file_name, 'w', encoding='utf-8'
                )

            if request.ignore_empty_fields:
                data = filter_data_fields(data, field_counts[resource_id])

            if not request.separate_files:
                # if the data is being written into one file we need to indicate which resource the
                # data came from, this is how we do that
                data['Source resource ID'] = resource_id

            # dump the data ensuring it works correctly as unicode
            open_files[resource_id].write(json.dumps(data, ensure_ascii=False))
            open_files[resource_id].write('\n')

        yield write
    finally:
        # make sure we close all the files we opened
        for open_file in open_files.values():
            open_file.close()



import contextlib
from collections import defaultdict
from pathlib import Path
from typing import List, Dict

from openpyxl import Workbook

from .sv import flatten_dict
from .utils import filter_data_fields, get_fields

RESOURCE_ID_FIELD_NAME = 'Source resource ID'
DEFAULT_SHEET_NAME = 'Data'
ALL_IN_ONE_FILE_NAME = 'data.xlsx'


class WorkbookWithFields:
    """
    Class to keep a Workbook and the associated fields in use within it together, plus
    some convenience methods for adding and saving data.
    """

    def __init__(self, fields: List[str]):
        '''
        :param fields: a list of field names
        '''
        self.fields = fields
        # the write only workbook avoids storing the workbook in memory and thus dodges running out
        # of RAM, see: https://openpyxl.readthedocs.io/en/stable/optimized.html#write-only-mode.
        # Note that there are some restrictions as to what you can do with a write only workbook but
        # we don't run into any of them in the current implementation and we need it so that this
        # writer doesn't eat all our RAM (as it says in the doc, RAM usage can be 50x file size! :O)
        self.workbook = Workbook(write_only=True)
        self.workbook.create_sheet(DEFAULT_SHEET_NAME)
        self.workbook.active.append(fields)

    def add(self, row: Dict[str, str]):
        """
        Add a row to the workbook's active sheet.

        :param row: the row of data as a dict
        """
        self.workbook.active.append(row.get(field) for field in self.fields)

    def save(self, path: Path):
        """
        Write the workbook to disk. This can only be called once (see write only
        workbook mode) and close the workbook.

        :param path: the path to write the workbook to
        """
        self.workbook.save(path)
        self.workbook.close()


@contextlib.contextmanager
def xlsx_writer(request, target_dir, field_counts):
    """
    An XLSX (i.e. Excel speadsheet) writer.

    :param request: the download request object
    :param target_dir: the directory to save the data to
    :param field_counts: a dict of resource ids -> fields -> counts used to determine which fields
                         should be included in the xlsx file headers
    :return: yields a write function
    """
    if request.separate_files:
        # map of resource ids to workbooks
        workbooks = {}
    else:
        all_field_names = get_fields(field_counts, request.ignore_empty_fields)
        # add the special resource id field name to the start
        all_field_names.insert(0, RESOURCE_ID_FIELD_NAME)
        workbook = WorkbookWithFields(all_field_names)
        # create a defaultdict which always returns the workbook we just created for all the data
        workbooks = defaultdict(lambda: workbook)

    try:

        def write(hit, data, resource_id):
            if request.separate_files and resource_id not in workbooks:
                field_names = get_fields(
                    field_counts, request.ignore_empty_fields, [resource_id]
                )
                workbooks[resource_id] = WorkbookWithFields(field_names)

            if request.ignore_empty_fields:
                data = filter_data_fields(data, field_counts[resource_id])

            row = flatten_dict(data)

            if not request.separate_files:
                # if the data is being written into one file we need to indicate which resource the
                # data came from, this is how we do that
                row['Source resource ID'] = resource_id

            workbooks[resource_id].add(row)

        yield write
    finally:
        target_path = Path(target_dir)
        if request.separate_files:
            # save each workbook out in turn
            for res_id, workbook_with_fields in workbooks.items():
                workbook_with_fields.save(target_path / f'{res_id}.xlsx')
        else:
            # there should only be one workbook in the workbooks dict so grab it and save it
            next(iter(workbooks.values())).save(target_path / ALL_IN_ONE_FILE_NAME)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-versioned-datastore
# Created by the Natural History Museum in London, UK
from ckan.plugins import toolkit

from . import datastore, search, downloads

blueprints = [datastore.blueprint, search.blueprint]

if toolkit.config.get(u'debug', False):
    blueprints.append(downloads.blueprint)



from collections import defaultdict

from splitgill.search import create_version_query
from elasticsearch_dsl import Search, A

from ..datastore_utils import prefix_resource, prefix_field, iter_data_fields


def get_fields(field_counts, ignore_empty_fields, resource_ids=None):
    """
    Return a sorted list of field names for the resource ids specified using the field
    counts dict. The field counts dict should provide the field names available and
    their counts at the given version for the given search, for each resource in the
    search. If ignore_empty_fields is True, then fields with a count of 0 will be
    ignored and not returned in the list.

    The list is sorted in ascending order using lowercase comparisons.

    :param field_counts: the dict of resource ids -> fields -> counts
    :param ignore_empty_fields: whether fields with no values should be included in the resulting
                                list or not
    :param resource_ids: the resource ids to get the fields for. The default is None which means
                         that the fields from all resources will be returned
    :return: a list of fields in case-insensitive ascending order
    """
    # TODO: retrieve the sort order for resources from the database and use
    field_names = set()

    for resource_id, counts in field_counts.items():
        if resource_ids is None or resource_id in resource_ids:
            for field, count in counts.items():
                if count == 0 and ignore_empty_fields:
                    continue
                else:
                    field_names.add(field)

    return sorted(field_names, key=lambda f: f.lower())


def calculate_field_counts(request, es_client):
    """
    Given a download request and an elasticsearch client to work with, work out the
    number of values available per field, per resource for the search.

    :param request: the DownloadRequest object
    :param es_client: the elasticsearch client to use
    :return: a dict of resource ids -> fields -> counts
    """
    field_counts = defaultdict(dict)
    for resource_id, version in request.resource_ids_and_versions.items():
        index_name = prefix_resource(resource_id)
        # get the base field mapping for the index so that we know which fields to look up, this
        # will get all fields from all versions and therefore isn't usable straight off the bat, we
        # have to then go and see which fields are present in the search at this version
        mapping = es_client.indices.get_mapping(index_name)[index_name]

        search = (
            Search.from_dict(request.search)
            .index(index_name)
            .using(es_client)
            .extra(size=0)
            .filter(create_version_query(version))
        )

        # get all the fields names and use dot notation for nested fields
        fields = ['.'.join(parts) for parts, _config in iter_data_fields(mapping)]
        for field in fields:
            # add a search which finds the documents that have a value for the given field at the
            # right version
            agg = A('value_count', field=prefix_field(field))
            search.aggs.bucket(field, agg)

        response = search.execute()
        for field in fields:
            field_counts[resource_id][field] = response.aggregations[field].value

    return field_counts


def filter_data_fields(data, field_counts, prefix=None):
    """
    Returns a new dict containing only the keys and values from the given data dict
    where the.

    corresponding field in the field_counts dict has a value greater than 0 - i.e. removes all
    fields from the data dict that shouldn't be included.

    Note that this may seem like a pointless exercise as surely if the field count is 0 for a field
    then it won't appear in any of the data dicts - however, because the data returned from
    elasticsearch is the source dict that was uploaded to it, it could contain nulls and empty
    string values. The calculate_field_counts function above that generates the field counts does it
    by using exist queries and because these count indexed values it skips nulls and empty strings.

    :param data: the data dict
    :param field_counts: a dict of field names and counts, the field names should be dot separated
                         for nested fields
    :param prefix: the prefix under which the fields in the passed data dict exist - this is used to
                   produce the field names for nested fields
    :return: a new dict containing only the fields from the original data dict that had a value
             other than 0 in the fields_count dict
    """
    filtered_data = {}
    for field, value in data.items():
        if prefix is not None:
            path = f'{prefix}.{field}'
        else:
            path = field

        # if the field contains a list of dicts we need to recurse for each one
        if isinstance(value, list) and value and isinstance(value[0], dict):
            filtered_value = []
            for element in value:
                filtered_element = filter_data_fields(
                    element, field_counts, prefix=path
                )
                # if there is any data left in the element after filtering, add it to the temp list
                if filtered_element:
                    filtered_value.append(filtered_element)
            # if there are any dicts left from the filtering, include them directly using the name
            # of the field, not the path. We don't need to check if the field has any values because
            # we know that it does because there are dicts left in the filtered list
            if filtered_value:
                filtered_data[field] = filtered_value
        # if the field is a dict, recurse to filter
        elif isinstance(value, dict):
            filtered_value = filter_data_fields(value, field_counts, prefix=path)
            # if there is any data left after the filtering, include the dict value directly, using
            # the name of the field, not the path. We don't need to check if the field has any
            # values because we know that it does because the filtered_value isn't empty
            if filtered_value:
                filtered_data[field] = filtered_value
        # for everything else, just check that the path is in the fields_count dict
        elif field_counts.get(path, 0):
            filtered_data[field] = value

    return filtered_data



from datetime import datetime

from ckan.plugins import toolkit
from splitgill.search import create_version_query
from splitgill.utils import to_timestamp
from elasticsearch_dsl import Search

from .meta import help, schema
from ckantools.decorators import action
from ...lib import common
from ...lib.basic_query.search import create_search
from ...lib.datastore_utils import (
    prefix_resource,
    is_datastore_resource,
    get_public_alias_name,
)
from ...lib.query.schema import get_latest_query_version


@action(
    schema.datastore_get_record_versions(), help.datastore_get_record_versions, get=True
)
def datastore_get_record_versions(id, resource_id):
    """
    Retrieves all the versions of the record with the given id in the given resource.
    The versions are returned as a list in ascending order.

    :param id: the id of the record
    :param resource_id: the id of the resource the record is in
    :return: a list of versions in ascending order
    """
    index_name = prefix_resource(resource_id)
    return common.SEARCH_HELPER.get_record_versions(index_name, int(id))


@action(schema.datastore_search(), help.datastore_get_resource_versions, get=True)
def datastore_get_resource_versions(
    resource_id, context, data_dict, original_data_dict
):
    """
    Retrieves all the versions of the given resource when under the given search. Note
    that the schema used for this action is the same as the datastore_search schema. The
    return is a dict including the version timestamp, the number of records modified in
    the version and the total records at the version.

    :param resource_id: the id of the resource to examine
    :param context: the context dict from the action call
    :param data_dict: the data_dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :return:
    """
    original_data_dict, data_dict, version, search = create_search(
        context, data_dict, original_data_dict
    )
    index_name = prefix_resource(resource_id)

    data = common.SEARCH_HELPER.get_index_version_counts(index_name, search=search)

    search = search.using(common.ES_CLIENT).index(index_name)[0:0]
    for result in data:
        version = result['version']
        count = search.filter(create_version_query(version)).count()
        result['count'] = count
    return data


@action(
    schema.datastore_get_rounded_version(), help.datastore_get_rounded_version, get=True
)
def datastore_get_rounded_version(resource_id, version=None):
    """
    Retrieve the closest version of the resource's data to the given version when
    rounding down.

    :param resource_id: the id of the resource
    :param version: the version timestamp. If None (the default) the latest version of the resource
                    is returned
    :return: the rounded version timestamp
    """
    index_name = prefix_resource(resource_id)
    return common.SEARCH_HELPER.get_rounded_versions([index_name], version)[index_name]


@action(
    schema.datastore_is_datastore_resource(),
    help.datastore_is_datastore_resource,
    get=True,
)
def datastore_is_datastore_resource(resource_id):
    """
    Checks whether the given resource id is in the datastore or not.

    :param resource_id: the resource to check
    :return: True if it is, False if not
    """
    return is_datastore_resource(resource_id)


@action({}, help.datastore_get_latest_query_schema_version, get=True)
def datastore_get_latest_query_schema_version():
    """
    Simply returns the latest available query schema version.

    :return: the query schema version
    """
    return get_latest_query_version()


@action(schema.datastore_count(), help.datastore_count, get=True)
def datastore_count(resource_ids=None, version=None):
    if version is None:
        version = to_timestamp(datetime.now())
    if resource_ids is None:
        resource_ids = ['*']

    indexes = [get_public_alias_name(resource_id) for resource_id in resource_ids]
    search = Search(using=common.ES_CLIENT, index=indexes).filter(
        create_version_query(version)
    )

    return search.count()



import logging

from ckan.plugins import toolkit

log = logging.getLogger(__name__)
base_url = toolkit.config.get('ckan.site_url')
object_endpoint = toolkit.config.get(
    'ckanext.versioned_datastore.record_view_endpoint', 'object.view'
)


class Transform:
    @classmethod
    def transform_data(cls, data, config):
        id_as_url = config.get('id_as_url')
        if id_as_url:
            field = id_as_url.get('field', 'id')
            data = cls.id_as_url(data, field)
        return data

    @classmethod
    def id_as_url(cls, data, field):
        """
        Reformat an ID field as a URL (probably one that links to that record). Requires
        an endpoint (config option ckanext.versioned_datastore.record_view_endpoint,
        default 'object.view') taking the ID field as the 'uuid' named argument.

        :param data: the record data to be transformed
        :param field: the name of the data field that contains the ID and that will contain the URL
        :return: the transformed data (or untransformed if there was an error).
        """
        try:
            object_id = data.get(field)
            if object_id is None or object_id == '':
                log.error(f'Failed to get uuid from field "{field}".')
                return data
            kwargs = {'uuid': object_id}
            url = toolkit.url_for(object_endpoint, **kwargs)
        except:
            log.error(f'Failed to generate URL from ID.', exc_info=True)
            return data
        data[field] = base_url + url
        return data



from datetime import datetime

from ckan.model import meta, DomainObject
from ckan.model.types import make_uuid
from sqlalchemy import Column, Table, BigInteger, UnicodeText, DateTime
from sqlalchemy.dialects.postgresql import JSONB

state_complete = 'complete'
state_failed = 'failed'
state_processing = 'processing'
state_zipping = 'zipping'

# this table stores query slugs
datastore_downloads_table = Table(
    'versioned_datastore_downloads',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('query_hash', UnicodeText, nullable=False, index=True),
    Column('query', JSONB, nullable=False),
    Column('query_version', UnicodeText, nullable=False),
    Column('resource_ids_and_versions', JSONB, nullable=False),
    Column('created', DateTime, nullable=False, default=datetime.utcnow),
    Column('total', BigInteger, nullable=True),
    Column('resource_totals', JSONB, nullable=True),
    Column('state', UnicodeText, nullable=True),
    Column('error', UnicodeText, nullable=True),
    Column('options', JSONB, nullable=True),
)


class DatastoreDownload(DomainObject):
    """
    Object for a datastore download row.
    """

    pass


meta.mapper(DatastoreDownload, datastore_downloads_table)



import contextlib
from collections import defaultdict

import os
import csv

from .utils import get_fields, filter_data_fields


def flatten_dict(data, path=None, separator=' | '):
    """
    Flattens a given dictionary so that nested dicts and lists of dicts are available
    from the root of the dict. For nested dicts, the keys in the nested dict are simply
    concatenated to the key that references the dict with a dot between each, for
    example:

        {"a": {"b": 4, "c": 6}} -> {"a.b": 4, "a.c": 6}

    This works to any nesting level.

    For lists of dicts, the common keys between them are pulled up to the level above in the same
    way as the standard nested dict, but if there are multiple dicts with the same keys the values
    associated with them are concatenated together using the separator parameter. For example:

        {"a": [{"b": 5}, {"b": 19}]} -> {"a.b": "5 | 19"}

    :param data: the dict to flatten
    :param path: the path to place all found keys under, by default this is None and therefore the
                 keys in the dict are not placed under anything and are used as is. This is really
                 only here for internal recursive purposes.
    :param separator: the string to use when concatenating lists of values, whether common ones from
                      a list of dicts, or indeed just a normal list of values
    :return: the flattened dict
    """
    flat = {}
    for key, value in data.items():
        if path is not None:
            # use a dot to indicate this key is below the parent in the path
            key = f'{path}.{key}'

        if isinstance(value, dict):
            # flatten the nested dict and then update the current dict we've got on the go
            flat.update(flatten_dict(value, path=key))
        elif isinstance(value, list):
            if all(isinstance(element, dict) for element in value):
                for element in value:
                    # iterate through the list of dicts flattening each as we go and then either
                    # just adding the value to the dict we've got on the go or appending it to the
                    # string value we're using for collecting multiples
                    for subkey, subvalue in flatten_dict(element, path=key).items():
                        if subkey not in flat:
                            flat[subkey] = subvalue
                        else:
                            flat[subkey] = f'{flat[subkey]}{separator}{subvalue}'
            else:
                flat[key] = separator.join(map(str, value))
        else:
            flat[key] = value

    return flat


def create_sv_writer(field_names, dialect, target_dir, filename):
    """
    Creates a file object and a reader that can be used to write data into the given
    file name and then returns both.

    :param field_names: the field names to include
    :param dialect: the python csv module dialect to use
    :param target_dir: the directory to write the data into
    :param filename: the filename to store the data in
    :return: a 2-tuple containing the open file and the DictWriter object
    """
    path = os.path.join(target_dir, filename)
    open_file = open(path, 'w', encoding='utf-8')
    writer = csv.DictWriter(open_file, field_names, dialect=dialect)
    writer.writeheader()
    return open_file, writer


@contextlib.contextmanager
def sv_writer(request, target_dir, field_counts, dialect, extension):
    """
    Provides the functionality to write data to *sv files (for example: csv and tsv).
    This function handles opening and closing the necessary files to write data to
    either one file or a file per resource.

    This function is a context manager and when used with `with` will yield a write function. This
    function takes 3 parameters - the elasticsearch hit object, the data dict from the hit and the
    resource id the hit came from.

    :param request: the download request object
    :param target_dir: the target directory to put the data files in
    :param field_counts: a dict of resource ids -> fields -> counts used to determine which fields
                         should be included in the *sv file headers
    :param dialect: the python csv dialect name to use
    :param extension: the extension to use on the file(s)
    :return: yields a write function
    """
    # we might be opening up a file for each resource, keep track so that we can close them after
    open_files = []

    if request.separate_files:
        # writers will be created lazily and stored in this dict using the resource ids as keys
        writers = {}
    else:
        # create a single writer object and open a single file for it
        all_field_names = get_fields(field_counts, request.ignore_empty_fields)
        # TODO: this could clash
        all_field_names = ['Source resource ID'] + all_field_names
        open_file, writer = create_sv_writer(
            all_field_names, dialect, target_dir, f'data.{extension}'
        )
        open_files.append(open_file)
        # sneekily use a defaultdict to return the same writer object regardless of the resource id
        writers = defaultdict(lambda: writer)

    try:

        def write(hit, data, resource_id):
            if request.separate_files and resource_id not in writers:
                # the file needs opening and the *sv writer object needs creating
                field_names = get_fields(
                    field_counts, request.ignore_empty_fields, [resource_id]
                )
                open_file, writer = create_sv_writer(
                    field_names, dialect, target_dir, f'{resource_id}.{extension}'
                )
                open_files.append(open_file)
                writers[resource_id] = writer

            if request.ignore_empty_fields:
                data = filter_data_fields(data, field_counts[resource_id])

            # the data dict may be nested, flatten it out first
            row = flatten_dict(data)
            if not request.separate_files:
                # if the rows are being written into one file we need to indicate which resource the
                # row came from, this is how we do that
                row['Source resource ID'] = resource_id

            # write the row to the writer, for separate files, this will select the correct writer
            # object and use it, for the single writer this will always retrieve the one writer
            # we're using and use it
            writers[resource_id].writerow(row)

        yield write
    finally:
        # make sure we close all the files we opened
        for open_file in open_files:
            open_file.close()



from ckan.plugins import interfaces


class IVersionedDatastore(interfaces.Interface):
    """
    Allow modifying versioned datastore logic.
    """

    def datastore_modify_data_dict(self, context, data_dict):
        """
        Allows modification of the data dict before it is validated and used to create
        the search object. This function should be used to remove/add/alter parameters
        in the data dict.

        :param context: the context
        :type context: dictionary
        :param data_dict: the parameters received from the user
        :type data_dict: dictionary
        """
        return data_dict

    def datastore_modify_search(self, context, original_data_dict, data_dict, search):
        """
        Allows modifications to the search before it is made. This is kind of analogous
        to IDatastore.datastore_search however instead of passing around a query dict,
        instead an elasticsearch-dsl Search object is passed.

        Each extension which implements this interface will be called in the order CKAN loaded them
        in, The search parameter will be the output of the previous extension's interface
        implementation, thus creating a chain of extensions, each getting a go at altering the
        search object if necessary. The base datastore_search function provides the initial search
        object.

        Implementors of this function should return the search object. Don't forget that most
        functions on the search object are chainable and create a copy of the search object - ensure
        you're returning the modified object!

        Two data dicts are passed into this function, the original data dict as it was before any
        ``datastore_modify_data_dict`` functions got to it, and the modified data_dict that was used
        to create the search object by the core functionality. This allows someone to, for example,
        remove a part of the data_dict in ``datastore_modify_data_dict`` to avoid it being added
        into the search object by the core functionality. Then, by implementing this function,
        they can add their custom search parts based on the details they removed by extracting them
        from the original_data_dict.

        :param context: the context
        :type context: dictionary
        :param original_data_dict: the parameters received from the user
        :type original_data_dict: dictionary
        :param data_dict: the parameters received from the user after they have been modified by
                          implementors of ``datastore_modify_data_dict``
        :type data_dict: dictionary
        :param search: the current search, as changed by the previous IVersionedDatastore extensions
                       in the chain
        :type search: elasticsearch-dsl Search object

        :returns: the search object with your modifications
        :rtype: elasticsearch-dsl Search object
        """
        return search

    def datastore_modify_result(self, context, original_data_dict, data_dict, result):
        """
        Allows modifications to the result after the search.

        Each extension which implements this interface will be called in the order CKAN loaded them
        in, The result parameter will be the output of the previous extension's interface
        implementation, thus creating a chain of extensions, each getting a go at altering the
        result object if necessary.

        Implementors of this function should return the result object so that the datastore_search
        function can build the final return dict.

        :param context: the context
        :type context: dictionary
        :param original_data_dict: the parameters received from the user
        :type original_data_dict: dictionary
        :param data_dict: the parameters received from the user after they have been modified by
                          implementors of ``datastore_modify_data_dict``
        :type data_dict: dictionary
        :param result: the current result, as changed by the previous IVersionedDatastore extensions
                       in the chain
        :type result: elasticsearch result object

        :returns: the result object with your modifications
        :rtype: elasticsearch result object
        """
        return result

    def datastore_modify_fields(self, resource_id, mapping, fields):
        """
        Allows modification of the field definitions before they are returned with the
        results of a datastore_search. The definitions are used in CKAN by the recline
        view and therefore need to abide by any of its requirements. By default all
        fields are included and are simply made up of a dict containing an id and type
        key. The id is the name of the field and the type is always string.

        :param resource_id: the resource id that was searched
        :param mapping: the mapping for the elasticsearch index containing the resource's data. This
                        is the raw mapping as a dict, retrieved straight from elasticsearch's
                        mapping endpoint
        :param fields: the field definitions that have so far been extracted from the mapping, by
                       default this is all fields
        :return: the list of field definition dicts
        """
        return fields

    def datastore_modify_index_doc(self, resource_id, index_doc):
        '''
        Action allowing the modification of a resource's data during indexing. The index_doc passed
        is a dict in the form:

            {
                "data": {},
                "meta": {}
            }

        which will be sent in this form to elasticsearch for indexing. The data key's value contains
        the data for the record at a version. The meta key's value contains metadata for the record
        so that we can search it correctly. Breakdown of the standard keys in the meta dict:

            - versions: a dict containing the range of versions this document is valid for. This is
                        represented using an elasticsearch range, with "gte" for the first valid
                        version and "lt" for the last version. If the "lt" key is missing the data
                        is current.
            - version: the version of this record this data represents
            - next_version: will be missing if this data is current but if present, this holds the
                            value of the next version of this record

        If needed, the record id will be located in the index_doc under the key '_id'.

        :param resource_id: the id of the resource being indexed
        :param index_doc: a dict that will be sent to elasticsearch
        :return: the dict for elasticsearch to index
        '''
        return index_doc

    def datastore_is_read_only_resource(self, resource_id):
        """
        Allows implementors to designate certain resources as read only. This is purely
        a datastore side concept and should be used to prevent actions such as:

            - creating a new datastore for the resource (i.e. creating the index in elasticsearch)
            - upserting data into the datastore for this resource
            - deleting the datastore for this resource
            - reindexing data in the datastore for this resource

        :param resource_id: the resource id to check
        :return: True if the resource should be treated as read only, False if not
        """
        return False

    def datastore_after_indexing(self, request, splitgill_stats, stats_id):
        """
        Allows implementors to hook onto the completion of an indexing task. This
        function doesn't return anything and any exceptions it raises will be caught and
        ignored.

        :param request: the ResourceIndexRequest object that triggered the indexing task
        :param splitgill_stats: the statistics about the indexing task from splitgill
        :param stats_id: the id of the statistics entry in the ImportStats database table
        """
        pass

    def datastore_reserve_slugs(self):
        """
        Allows implementors to reserve queries using reserved pretty slugs. Implementors
        should return a dict made up of reserved pretty slugs as keys and then the slug
        parameters as the values. These values should be another dict containing the
        following optional keys:

            - query, a dict query (defaults to {})
            - query_version, the query schema version (defaults to the latest query schema version)
            - version, the version of the data to search at (defaults to None)
            - resource_ids, a list of resource ids to search (defaults to all resource ids)
            - resource_ids_and_versions, a dict of resource ids and specific versions to search at
                                         (defaults to an empty dict)

        If a slug already exists in the database with the same reserved pretty slug and the same
        query parameters then nothing happens.

        If a slug already exists in the database with the same reserved pretty slug but a different
        set of query parameters then a DuplicateSlugException is raised.

        If a slug already exists in the database with the same query parameters but no reserved
        pretty slug then the reserved pretty slug is added to the slug.
        """
        return {}

    def datastore_modify_guess_fields(self, resource_ids, fields):
        """
        Allows plugins to manipulate the Fields object used to figure out the groups
        that should be returned by the datastore_guess_fields action.

        :param resource_ids: a list of resource ids
        :param fields: a Fields object
        :return: the Fields object
        """
        return fields

    def datastore_multisearch_modify_response(self, response):
        """
        Allows plugins to alter the response dict returned from the
        datastore_multisearch action before it is returned.

        :param response: the dict to be returned to the caller
        :return: the response dict
        """
        return response


class IVersionedDatastoreQuerySchema(interfaces.Interface):
    def get_query_schemas(self):
        """
        Hook to allow registering custom query schemas.

        :return: a list of tuples of the format (query schema version, schema object) where the
                 query schema version is a string of format v#.#.# and the schema object is an
                 instance of ckanext.versioned_datastore.lib.query.Schema
        """
        return []


class IVersionedDatastoreDownloads(interfaces.Interface):
    def download_modify_email_templates(self, text_template: str, html_template: str):
        """
        Hook allowing other extensions to modify the templates used when sending
        download emails. The templates can be modified in place or completely replaced.

        :param text_template: the text email template string
        :param html_template: the html email template string
        :return: a 2-tuple containing the text template string and the html template string
        """
        return text_template, html_template

    def download_modify_email_template_context(self, request, context):
        """
        Hook allowing other extensions to modify the templating context used to generate
        the download email (both plain text and HTML versions) before it is sent.

        The default context contains:
            - "download_url": the download zip's full URL
            - "site_url": the CKAN site's full URL (this is taken straight from the config)

        :param request: the DownloadRequest object
        :param context: templating context dict
        :return: context templating dict
        """
        return context

    def download_before_write(self, request):
        """
        Hook allowing other extensions to access and modify the request before the
        search is run and the download file is created.

        :param request: the DownloadRequest object
        :return: the request
        """
        return request



import csv
import json
import os
import shutil
from collections import OrderedDict
from datetime import datetime as dt

from ckan.plugins import plugin_loaded, toolkit
from lxml import etree

from .schema import Schema
from .urls import TDWGUrls, XMLUrls, GBIFUrls
from .utils import NSMap, json_to_xml, get_setting


class Archive(object):
    def __init__(self, schema: Schema, request, target_dir, output_name=None):
        """
        Produces one output file. Each archive contains a core .csv data file, one .csv
        file for each specified extension, a metafile (meta.xml), and a resource
        metadata file (eml.xml).

        :param schema: a Schema object representing the output data structure
        :param request: the request object
        :param target_dir: the root folder to save files into
        :param output_name: only required if resources are to be separated into different files;
                            usually this is the resource id
        """
        self.schema = schema
        self.dir = target_dir
        self.request = request
        if request.separate_files and not output_name:
            raise Exception(
                'output_name must be specified if separate files are required.'
            )
        self.output_name = output_name or 'dwca'
        self._core_file = None
        self._core_writer = None
        self._ext_files = None
        self._ext_writers = None
        self.rows_written = 0

    @property
    def _uid(self):
        """
        A unique identifier for this particular archive.
        """
        if self.single_resource:
            return self.request.resource_ids[0]
        elif self.request.separate_files:
            return self.output_name
        else:
            return self.request.query_hash

    @property
    def _build_dir(self):
        return os.path.join(self.dir, self._uid)

    @property
    def _core_file_name(self):
        return f'{self.schema.row_type_name.lower()}.csv'

    @property
    def single_resource(self):
        """
        If this download will only include data from a single resource with no filters,
        i.e. just the whole resource.
        """
        return len(self.request.resource_ids) == 1 and self.request.query == {}

    def open(self, field_names):
        """
        Create the build dir, open the necessary files, and create csv writers. Does.

        *not* write the csv headers; this is done in self.initialise().

        :param field_names: field names available in the data rows
        :return: self
        """
        os.mkdir(self._build_dir)
        root_field_names = set([f.split('.')[0] for f in field_names])
        all_ext_field_names = [
            item for ext in self.schema.extensions for item in ext.location.fields
        ]

        standard_fields = ['datasetID', 'basisOfRecord', 'dynamicProperties']
        dataset_fields = [
            f
            for f in root_field_names
            if f not in all_ext_field_names and f in self.schema.props
        ]
        core_field_names = ['_id'] + list(set(standard_fields + dataset_fields))
        self._core_file = open(os.path.join(self._build_dir, self._core_file_name), 'w')
        self._core_writer = csv.DictWriter(
            self._core_file, core_field_names, dialect='unix'
        )

        self._ext_files = {}
        self._ext_writers = {}
        for e in self.schema.extensions:
            subfields = [
                f.split('.')[-1]
                for f in field_names
                if f.split('.')[0] in e.location.fields
            ]
            potential_fields = self.schema.extension_props[e.name]
            ext_field_names = ['_id'] + list(
                set([f for f in subfields if f in potential_fields])
            )
            open_file = open(
                os.path.join(self._build_dir, f'{e.name.lower()}.csv'), 'w'
            )
            self._ext_files[e.name] = open_file
            writer = csv.DictWriter(open_file, ext_field_names, dialect='unix')
            self._ext_writers[e.name] = writer
        return self

    def close(self):
        """
        Close any open files, build the metadata files, and zip everything up.
        """
        if self._core_file is None:
            # this might be called several times due to the defaultdict implementation in writer.py
            return
        self._core_file.close()
        for f in self._ext_files.values():
            f.close()
        with open(os.path.join(self._build_dir, 'meta.xml'), 'w') as f:
            meta_content = self.make_meta()
            f.write(meta_content)
        with open(os.path.join(self._build_dir, 'eml.xml'), 'w') as f:
            eml_content = self.make_eml()
            f.write(eml_content)
        shutil.make_archive(
            os.path.join(self.dir, self.output_name), 'zip', self._build_dir
        )
        shutil.rmtree(self._build_dir)
        self._core_file = None
        self._core_writer = None
        self._ext_files = None
        self._ext_writers = None

    def initialise(self, record):
        """
        Only runs for the first row in the dataset. Checks that certain fields are the
        correct data type before writing the headers.

        :param record: a single row
        """
        if self.rows_written > 0:
            return
        # check 'type' is right
        valid_types = [
            'StillImage',
            'MovingImage',
            'Sound',
            'PhysicalObject',
            'Event',
            'Text',
        ]
        self._core_writer.fieldnames = sorted(self._core_writer.fieldnames)
        if 'type' in record and record['type'] not in valid_types:
            self._core_writer.fieldnames = [
                f for f in self._core_writer.fieldnames if f != 'type'
            ]
        self._core_writer.writeheader()
        for writer in self._ext_writers.values():
            writer.writeheader()

    def _extract_record(self, record, id_field):
        """
        Transform the data in a single row into the required format. Separates extension
        fields from core fields and filters out any fields that don't match the schema.

        :param record: the row of data
        :param id_field: the name of the current id field
        :return: core row (dict), extension rows keyed on extension name (dict of lists of dicts)
        """
        core = {}
        ext = {}
        dynamic_properties = {}

        if id_field not in record:
            raise Exception(f'Record does not have ID field {id_field}')
        record_id = record.get(id_field)
        core['_id'] = record_id

        extension_map = {}
        for e in self.schema.extensions:
            for f in e.location.fields:
                extension_map[f] = e.name

        for k, v in record.items():
            if k in extension_map:
                ext_props = self._ext_writers[extension_map[k]].fieldnames

                def _extract_ext(subdict):
                    props = {'_id': record_id}
                    for ek, ev in subdict.items():
                        if ek in ext_props:
                            props[ek] = ev
                    return props

                if isinstance(v, list):
                    ext_extracted = [_extract_ext(x) for x in v]
                elif isinstance(v, dict):
                    ext_extracted = [_extract_ext(v)]
                else:
                    ext_extracted = [_extract_ext({k: v})]
                ext[extension_map[k]] = ext_extracted
            else:
                if k in self._core_writer.fieldnames:
                    core[k] = v
                else:
                    dynamic_properties[k] = v

        core['dynamicProperties'] = json.dumps(dynamic_properties)
        return core, ext

    def write_record(self, record, id_field='_id'):
        """
        Transform the record then write it to csv.

        :param record: the row of data
        :param id_field: the name of the current id field
        """
        if self._core_file is None:
            # if this is None, the file is not open, and this is being called from the wrong place.
            raise Exception('File is not open.')
        core_row, ext_rows = self._extract_record(record, id_field)
        self._core_writer.writerow(core_row)
        self.rows_written += 1
        for e, rows in ext_rows.items():
            for row in rows:
                self._ext_writers[e].writerow(row)

    def make_meta(self):
        """
        Create the xml text content of the metafile.

        :return: xml string
        """
        nsmap = NSMap(xsi=XMLUrls.xsi, xs=XMLUrls.xs)
        root = etree.Element(
            'archive',
            nsmap=nsmap,
            attrib={
                'xmlns': TDWGUrls.xmlns,
                'metadata': 'eml.xml',
                nsmap.ns('xsi', 'schemaLocation'): ' '.join(
                    [TDWGUrls.xmlns, TDWGUrls.metadata]
                ),
            },
        )
        attributes = {
            'encoding': 'UTF-8',
            'linesTerminatedBy': '\n',
            'fieldsTerminatedBy': ',',
            'fieldsEnclosedBy': '"',
            'ignoreHeaderLines': '1',
        }
        core = etree.SubElement(
            root, 'core', rowType=self.schema.row_type, **attributes
        )
        core_files = etree.SubElement(core, 'files')
        core_files_location = etree.SubElement(core_files, 'location')
        core_files_location.text = self._core_file_name
        etree.SubElement(core, 'id', index='0')
        for i, c in enumerate(self._core_writer.fieldnames):
            if c == '_id':
                continue
            etree.SubElement(
                core, 'field', index=str(i), term=self.schema.props.get(c).iri
            )
        for e in self.schema.extensions:
            ext_root = etree.SubElement(
                root, 'extension', rowType=e.row_type, **attributes
            )
            ext_files = etree.SubElement(ext_root, 'files')
            ext_files_location = etree.SubElement(ext_files, 'location')
            ext_files_location.text = f'{e.name.lower()}.csv'
            etree.SubElement(ext_root, 'coreid', index='0')
            ext_props = self.schema.extension_props[e.name]
            for i, c in enumerate(self._ext_writers[e.name].fieldnames):
                if c == '_id':
                    continue
                etree.SubElement(
                    ext_root, 'field', index=str(i), term=ext_props.get(c).iri
                )
        return etree.tostring(root, pretty_print=True).decode()

    def make_eml(self):
        """
        Create the xml text content of the resource metadata file.

        Tries to use some sensible
        defaults and get information from other relevant plugins where available, but there's still
        the potential for errors or silly data.
        :return: xml string
        """
        # load some useful actions so we don't have to fetch them repeatedly
        resource_show = toolkit.get_action('resource_show')
        package_show = toolkit.get_action('package_show')

        # get the resources and packages associated with the request
        resources = [resource_show({}, {'id': r}) for r in self.request.resource_ids]
        packages = [package_show({}, {'id': r['package_id']}) for r in resources]

        # define some variables
        org = {
            'organizationName': get_setting(
                'ckanext.versioned_datastore.dwc_org_name',
                'ckanext.doi.publisher',
                'ckan.site_title',
            ),
            'electronicMailAddress': get_setting(
                'ckanext.versioned_datastore.dwc_org_email', 'smtp.mail_from'
            ),
            'onlineUrl': get_setting('ckan.site_url'),
        }
        site_name = get_setting('ckanext.doi.site_title', 'ckan.site_title')
        site_url = get_setting('ckan.site_url', default='')
        licenses = list(
            set(
                [
                    p.get('license_title')
                    for p in packages
                    if p.get('license_title') is not None
                ]
            )
        )
        query_license = (
            licenses[0]
            if len(licenses) == 1
            else get_setting(
                'ckanext.versioned_datastore.dwc_default_license', default='null'
            )
        )

        # set up the metadata dict in order (otherwise GBIF complains) with some defaults
        dataset_metadata = OrderedDict(
            {
                'alternateIdentifier': [self.request.query_hash],
                'title': f'Query on {site_name}',
                'creator': [org],
                'metadataProvider': [org],
                'pubDate': dt.now().strftime('%Y-%m-%d'),
                'language': get_setting('ckan.locale_default'),
                'abstract': {
                    'para': f'Query ID {self.request.query_hash} on {site_name} ({self.rows_written} records).'
                },
                'keywordSet': [],
                'intellectualRights': {'para': query_license},
                'distribution': ({'online': {'url': site_url}}, {'scope': 'document'}),
                'contact': [org],
            }
        )
        additional_metadata = {
            'metadata': {
                'gbif': {
                    'dateStamp': dt.now().strftime('%Y-%m-%dT%H:%M:%SZ'),
                    'citation': None,
                    'resourceLogoUrl': site_url
                    + toolkit.config.get('ckan.site_logo', ''),
                }
            }
        }

        if self.single_resource:
            # if it's just downloading one resource with no filters
            resource = resources[0]
            package = packages[0]
            dataset_metadata['title'] = resource['name']
            dataset_metadata['alternateIdentifier'] += self.request.resource_ids
            dataset_metadata['pubDate'] = package.get(
                'metadata_modified', resource.get('created')
            )
            dataset_metadata['abstract'] = {'para': resource.get('description')}
            dataset_metadata['distribution'][0]['online']['url'] = (
                site_url
                + toolkit.url_for(
                    'resource.read', id=package['id'], resource_id=resource['id']
                ),
                {'function': 'information'},
            )
            if 'doi' in package:
                dataset_metadata['alternateIdentifier'].append('doi:' + package['doi'])
                pkg_year = toolkit.h.package_get_year(package)
                pkg_title = package['title']
                pkg_doi = package['doi']
                pkg_author = package['author']
                site_title = toolkit.h.get_site_title()
                additional_metadata['metadata']['gbif']['citation'] = (
                    f'{pkg_author} ({pkg_year}). Dataset: {pkg_title}. {site_title}',
                    {'identifier': f'https://doi.org/{pkg_doi}'},
                )
        else:
            if plugin_loaded('query_dois'):
                from ckanext.query_dois.lib.doi import find_existing_doi

                query_doi = find_existing_doi(
                    self.request.resource_ids_and_versions,
                    self.request.query_hash,
                    self.request.query_version,
                )
                if query_doi:
                    dataset_metadata['alternateIdentifier'] += ['doi:' + query_doi.doi]
                    dataset_metadata['distribution'][0]['online']['url'] = (
                        site_url
                        + toolkit.url_for(
                            'query_doi.landing_page',
                            data_centre=get_setting('ckanext.query_dois.prefix'),
                            identifier=query_doi.doi,
                        ),
                        {'function': 'information'},
                    )
                    additional_metadata['metadata']['gbif']['citation'] = (
                        toolkit.h.create_multisearch_citation_text(
                            query_doi, html=False
                        ),
                        {'identifier': f'https://doi.org/{query_doi.doi}'},
                    )

        if (
            self.schema.core_extension
            and self.schema.core_extension.location.base == GBIFUrls.base_url
        ):
            dataset_metadata['keywordSet'].append(
                {
                    'keyword': self.schema.row_type_name,
                    'keywordThesaurus': f'GBIF Dataset Type Vocabulary: {GBIFUrls.thesaurus}',
                }
            )

        if plugin_loaded('attribution'):
            package_contributions_show = toolkit.get_action(
                'package_contributions_show'
            )
            authors = []
            for r in resources:
                contributions = package_contributions_show({}, {'id': r['package_id']})
                for c in contributions['contributions']:
                    if c['agent']['id'] in authors:
                        continue
                    agent = {}
                    if c['agent']['external_id']:
                        agent['userId'] = (
                            c['agent']['external_id'],
                            {'directory': c['agent']['external_id_url']},
                        )
                    if c['agent']['agent_type'] == 'person':
                        agent['individualName'] = {
                            'givenName': c['agent']['given_names'],
                            'surName': c['agent']['family_name'],
                        }
                    else:
                        agent['organizationName'] = c['agent']['name']
                    dataset_metadata['creator'].append(agent)
                    authors.append(c['agent']['id'])

        nsmap = NSMap(eml=XMLUrls.eml, dc=XMLUrls.dc, xsi=XMLUrls.xsi, xml=XMLUrls.xml)

        root = etree.Element(
            nsmap.ns('eml', 'eml'),
            nsmap=nsmap,
            attrib={
                'scope': 'system',
                'system': 'http://gbif.org',
                nsmap.ns('xsi', 'schemaLocation'): ' '.join(
                    [XMLUrls.eml, GBIFUrls.eml]
                ),
                'packageId': self._uid,
                nsmap.ns('xml', 'lang'): 'en',
            },
        )

        for i in json_to_xml('dataset', dataset_metadata):
            root.append(i)

        if additional_metadata['metadata']['gbif'].get('citation') is not None:
            for i in json_to_xml('additionalMetadata', additional_metadata):
                root.append(i)

        return etree.tostring(root, pretty_print=True).decode()
